import os
import tempfile
import unittest

import pytest

import cupy
from cupy import cuda
from cupy import testing
import numpy


class TestFromData(unittest.TestCase):

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_array(self, xp, dtype, order):
        return xp.array([[1, 2, 3], [2, 3, 4]], dtype=dtype, order=order)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_array_from_empty_list(self, xp, dtype, order):
        return xp.array([], dtype=dtype, order=order)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_array_from_nested_empty_list(self, xp, dtype, order):
        return xp.array([[], []], dtype=dtype, order=order)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_array_from_numpy(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), numpy, dtype)
        return xp.array(a, order=order)

    def test_array_from_numpy_blocking(self):
        a = testing.shaped_arange((2, 3, 4), numpy, numpy.float32)
        testing.assert_array_equal(cupy.array(a, blocking=True), a)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_array_from_numpy_scalar(self, xp, dtype, order):
        a = numpy.array(2, dtype=dtype)
        return xp.array(a, order=order)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal(
        # when zero-copy, we may not be able to guarantee strides
        strides_check=not bool(int(os.environ.get('CUPY_ENABLE_UMP', 0))))
    def test_array_from_numpy_c_and_f(self, xp, dtype, order):
        a = numpy.ones((1, 3, 1), dtype=dtype)
        return xp.array(a, order=order)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_array_from_numpy_broad_cast(self, xp, dtype, order):
        a = testing.shaped_arange((2, 1, 4), numpy, dtype)
        a = numpy.broadcast_to(a, (2, 3, 4))
        return xp.array(a, order=order)

    @testing.for_orders('CFAK', name='src_order')
    @testing.for_orders('CFAK', name='dst_order')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal(strides_check=True)
    def test_array_from_list_of_numpy(self, xp, dtype, src_order, dst_order):
        # compares numpy.array(<list of numpy.ndarray>) with
        # cupy.array(<list of numpy.ndarray>)
        a = [
            testing.shaped_arange((3, 4), numpy, dtype, src_order) + (12 * i)
            for i in range(2)]
        return xp.array(a, order=dst_order)

    @testing.for_orders('CFAK', name='src_order')
    @testing.for_orders('CFAK', name='dst_order')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal(strides_check=True)
    def test_array_from_list_of_numpy_view(self, xp, dtype, src_order,
                                           dst_order):
        # compares numpy.array(<list of numpy.ndarray>) with
        # cupy.array(<list of numpy.ndarray>)

        # create a list of view of ndarrays
        a = [
            (testing.shaped_arange((3, 8), numpy,
                                   dtype, src_order) + (24 * i))[:, ::2]
            for i in range(2)]
        return xp.array(a, order=dst_order)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal(strides_check=True)
    def test_array_from_list_of_numpy_scalar(self, xp, dtype, order):
        # compares numpy.array(<list of numpy.ndarray>) with
        # cupy.array(<list of numpy.ndarray>)
        a = [numpy.array(i, dtype=dtype) for i in range(2)]
        return xp.array(a, order=order)

    @testing.for_orders('CFAK', name='src_order')
    @testing.for_orders('CFAK', name='dst_order')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal(strides_check=True)
    def test_array_from_nested_list_of_numpy(self, xp, dtype, src_order,
                                             dst_order):
        # compares numpy.array(<list of numpy.ndarray>) with
        # cupy.array(<list of numpy.ndarray>)
        a = [
            [testing.shaped_arange(
                (3, 4), numpy, dtype, src_order) + (12 * i)]
            for i in range(2)]
        return xp.array(a, order=dst_order)

    @testing.for_orders('CFAK', name='src_order')
    @testing.for_orders('CFAK', name='dst_order')
    @testing.for_all_dtypes_combination(names=('dtype1', 'dtype2'))
    @testing.numpy_cupy_array_equal(strides_check=True)
    def test_array_from_list_of_cupy(
            self, xp, dtype1, dtype2, src_order, dst_order):
        # compares numpy.array(<list of numpy.ndarray>) with
        # cupy.array(<list of cupy.ndarray>)
        a = [
            testing.shaped_arange((3, 4), xp, dtype1, src_order),
            testing.shaped_arange((3, 4), xp, dtype2, src_order),
        ]
        return xp.array(a, order=dst_order)

    @testing.for_orders('CFAK', name='src_order')
    @testing.for_orders('CFAK', name='dst_order')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal(strides_check=True)
    def test_array_from_list_of_cupy_view(self, xp, dtype, src_order,
                                          dst_order):
        # compares numpy.array(<list of numpy.ndarray>) with
        # cupy.array(<list of cupy.ndarray>)

        # create a list of view of ndarrays
        a = [
            (testing.shaped_arange((3, 8), xp,
                                   dtype, src_order) + (24 * i))[:, ::2]
            for i in range(2)]
        return xp.array(a, order=dst_order)

    @testing.for_orders('CFAK', name='src_order')
    @testing.for_orders('CFAK', name='dst_order')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal(strides_check=True)
    def test_array_from_nested_list_of_cupy(self, xp, dtype, src_order,
                                            dst_order):
        # compares numpy.array(<list of numpy.ndarray>) with
        # cupy.array(<list of cupy.ndarray>)
        a = [
            [testing.shaped_arange((3, 4), xp, dtype, src_order) + (12 * i)]
            for i in range(2)]
        return xp.array(a, order=dst_order)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal(strides_check=True)
    def test_array_from_list_of_cupy_scalar(self, xp, dtype, order):
        # compares numpy.array(<list of numpy.ndarray>) with
        # cupy.array(<list of cupy.ndarray>)
        a = [xp.array(i, dtype=dtype) for i in range(2)]
        return xp.array(a, order=order)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_array_from_nested_list_of_cupy_scalar(self, xp, dtype, order):
        # compares numpy.array(<list of numpy.ndarray>) with
        # cupy.array(<list of cupy.ndarray>)
        a = [[xp.array(i, dtype=dtype)] for i in range(2)]
        return xp.array(a, order=order)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_array_copy(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        return xp.array(a, order=order)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal(strides_check=True)
    def test_array_both_c_and_f_contig_copy(self, xp, dtype, order):
        a = testing.shaped_arange((1, 4, 1), xp, dtype, order="C")
        return xp.array(a, order=order)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal(strides_check=True)
    def test_array_both_c_and_f_contig_f_strides_copy(self, xp, dtype, order):
        a = testing.shaped_arange((1, 4, 1), xp, dtype, order="F")
        return xp.array(a, order=order)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_array_copy_is_copied(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        b = xp.array(a, order=order)
        a.fill(0)
        return b

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes(name='dtype1', no_complex=True)
    @testing.for_all_dtypes(name='dtype2')
    @testing.numpy_cupy_array_equal()
    def test_array_copy_with_dtype(self, xp, dtype1, dtype2, order):
        # complex to real makes no sense
        a = testing.shaped_arange((2, 3, 4), xp, dtype1)
        return xp.array(a, dtype=dtype2, order=order)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes(name='dtype1', no_complex=True)
    @testing.for_all_dtypes(name='dtype2')
    @testing.numpy_cupy_array_equal()
    def test_array_copy_with_dtype_char(self, xp, dtype1, dtype2, order):
        # complex to real makes no sense
        a = testing.shaped_arange((2, 3, 4), xp, dtype1)
        return xp.array(a, dtype=numpy.dtype(dtype2).char, order=order)

    @testing.for_orders('CFAK')
    @testing.numpy_cupy_array_equal()
    def test_array_copy_with_dtype_being_none(self, xp, order):
        a = testing.shaped_arange((2, 3, 4), xp)
        return xp.array(a, dtype=None, order=order)

    @testing.for_orders('CFAK', name='src_order')
    @testing.for_orders('CFAK', name='dst_order')
    @testing.for_all_dtypes(name='dtype1', no_complex=True)
    @testing.for_all_dtypes(name='dtype2')
    @testing.numpy_cupy_array_equal(strides_check=True)
    def test_array_copy_list_of_numpy_with_dtype(self, xp, dtype1, dtype2,
                                                 src_order, dst_order):
        # compares numpy.array(<list of numpy.ndarray>) with
        # cupy.array(<list of numpy.ndarray>)
        a = [
            testing.shaped_arange((3, 4), numpy, dtype1, src_order) + (12 * i)
            for i in range(2)]
        return xp.array(a, dtype=dtype2, order=dst_order)

    @testing.for_orders('CFAK', name='src_order')
    @testing.for_orders('CFAK', name='dst_order')
    @testing.for_all_dtypes(name='dtype1', no_complex=True)
    @testing.for_all_dtypes(name='dtype2')
    @testing.numpy_cupy_array_equal(strides_check=True)
    def test_array_copy_list_of_numpy_with_dtype_char(self, xp, dtype1,
                                                      dtype2, src_order,
                                                      dst_order):
        # compares numpy.array(<list of numpy.ndarray>) with
        # cupy.array(<list of numpy.ndarray>)
        a = [
            testing.shaped_arange((3, 4), numpy, dtype1, src_order) + (12 * i)
            for i in range(2)]
        return xp.array(a, dtype=numpy.dtype(dtype2).char, order=dst_order)

    @testing.for_orders('CFAK', name='src_order')
    @testing.for_orders('CFAK', name='dst_order')
    @testing.for_all_dtypes(name='dtype1', no_complex=True)
    @testing.for_all_dtypes(name='dtype2')
    @testing.numpy_cupy_array_equal(strides_check=True)
    def test_array_copy_list_of_cupy_with_dtype(self, xp, dtype1, dtype2,
                                                src_order, dst_order):
        # compares numpy.array(<list of numpy.ndarray>) with
        # cupy.array(<list of cupy.ndarray>)
        a = [
            testing.shaped_arange((3, 4), xp, dtype1, src_order) + (12 * i)
            for i in range(2)]
        return xp.array(a, dtype=dtype2, order=dst_order)

    @testing.for_orders('CFAK', name='src_order')
    @testing.for_orders('CFAK', name='dst_order')
    @testing.for_all_dtypes(name='dtype1', no_complex=True)
    @testing.for_all_dtypes(name='dtype2')
    @testing.numpy_cupy_array_equal(strides_check=True)
    def test_array_copy_list_of_cupy_with_dtype_char(self, xp, dtype1, dtype2,
                                                     src_order, dst_order):
        # compares numpy.array(<list of numpy.ndarray>) with
        # cupy.array(<list of cupy.ndarray>)
        a = [
            testing.shaped_arange((3, 4), xp, dtype1, src_order) + (12 * i)
            for i in range(2)]
        return xp.array(a, dtype=numpy.dtype(dtype2).char, order=dst_order)

    @testing.with_requires('numpy<2.0')
    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_array_no_copy(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        b = xp.array(a, copy=False, order=order)
        a.fill(0)
        return b

    @testing.with_requires('numpy<2.0')
    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_array_f_contiguous_input(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), xp, dtype, order='F')
        b = xp.array(a, copy=False, order=order)
        return b

    @testing.with_requires('numpy<2.0')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_array_f_contiguous_output(self, xp, dtype):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        b = xp.array(a, copy=False, order='F')
        assert b.flags.f_contiguous
        return b

    @testing.multi_gpu(2)
    def test_array_multi_device(self):
        with cuda.Device(0):
            x = testing.shaped_arange((2, 3, 4), cupy, dtype='f')
        with cuda.Device(1):
            y = cupy.array(x)
        assert isinstance(y, cupy.ndarray)
        assert x is not y  # Do copy
        assert int(x.device) == 0
        assert int(y.device) == 1
        testing.assert_array_equal(x, y)

    @testing.multi_gpu(2)
    def test_array_multi_device_zero_size(self):
        with cuda.Device(0):
            x = testing.shaped_arange((0,), cupy, dtype='f')
        with cuda.Device(1):
            y = cupy.array(x)
        assert isinstance(y, cupy.ndarray)
        assert x is not y  # Do copy
        assert x.device.id == 0
        assert y.device.id == 1
        testing.assert_array_equal(x, y)

    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_array_no_copy_ndmin(self, xp, dtype):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        b = xp.array(a, copy=False, ndmin=5)
        assert a.shape == (2, 3, 4)
        a.fill(0)
        return b

    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_array_from_big_endian(self, xp, dtype):
        dtype = numpy.dtype(dtype).newbyteorder('>')
        a = testing.shaped_arange((2, 3, 4), numpy, dtype)
        b = xp.array(a)
        # Make a computation here as just moving big-endian data back and forth
        # happens to work before the change in #5828
        return b + b

    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_array_from_list_of_numpy_big_endian(self, xp, dtype):
        dtype = numpy.dtype(dtype).newbyteorder('>')
        a = [testing.shaped_arange((3, 4), numpy, dtype) for i in range(2)]
        b = xp.array(a)
        # Make a computation here as just moving big-endian data back and forth
        # happens to work before the change in #5828
        return b + b

    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_asarray(self, xp, dtype):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        return xp.asarray(a)

    def test_asarray_blocking(self):
        a = testing.shaped_arange((2, 3, 4), numpy, numpy.float32)
        testing.assert_array_equal(cupy.asarray(a, blocking=True), a)

    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_asarray_is_not_copied(self, xp, dtype):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        b = xp.asarray(a)
        a.fill(0)
        return b

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_asarray_with_order(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        b = xp.asarray(a, order=order)
        if order in ['F', 'f']:
            assert b.flags.f_contiguous
        else:
            assert b.flags.c_contiguous
        return b

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_asarray_preserves_numpy_array_order(self, xp, dtype, order):
        a_numpy = testing.shaped_arange((2, 3, 4), numpy, dtype, order)
        b = xp.asarray(a_numpy)
        assert b.flags.f_contiguous == a_numpy.flags.f_contiguous
        assert b.flags.c_contiguous == a_numpy.flags.c_contiguous
        return b

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_asanyarray_with_order(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        b = xp.asanyarray(a, order=order)
        if order in ['F', 'f']:
            assert b.flags.f_contiguous
        else:
            assert b.flags.c_contiguous
        return b

    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_asanyarray_from_big_endian(self, xp, dtype):
        dtype = numpy.dtype(dtype).newbyteorder('>')
        a = testing.shaped_arange((2, 3, 4), numpy, dtype)
        b = xp.asanyarray(a)
        # Make a computation here as just moving big-endian data back and forth
        # happens to work before the change in #5828
        return b + b

    def test_asanyarray_blocking(self):
        a = testing.shaped_arange((2, 3, 4), numpy, numpy.float32)
        testing.assert_array_equal(cupy.asanyarray(a, blocking=True), a)

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_asarray_from_numpy(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), numpy, dtype)
        b = xp.asarray(a, order=order)
        if order in ['F', 'f']:
            assert b.flags.f_contiguous
        else:
            assert b.flags.c_contiguous
        return b

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_asarray_with_order_copy_behavior(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        b = xp.asarray(a, order=order)
        a.fill(0)
        return b

    def test_ascontiguousarray_on_noncontiguous_array(self):
        a = testing.shaped_arange((2, 3, 4))
        b = a.transpose(2, 0, 1)
        c = cupy.ascontiguousarray(b)
        assert c.flags.c_contiguous
        testing.assert_array_equal(b, c)

    def test_ascontiguousarray_on_contiguous_array(self):
        a = testing.shaped_arange((2, 3, 4))
        b = cupy.ascontiguousarray(a)
        assert a is b

    @testing.numpy_cupy_array_equal()
    def test_asarray_cuda_array_zero_dim(self, xp):
        a = xp.ones(())
        return xp.ascontiguousarray(a)

    @testing.numpy_cupy_array_equal()
    def test_asarray_cuda_array_zero_dim_dtype(self, xp):
        a = xp.ones((), dtype=numpy.float64)
        return xp.ascontiguousarray(a, dtype=numpy.int64)

    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_asarray_from_big_endian(self, xp, dtype):
        dtype = numpy.dtype(dtype).newbyteorder('>')
        a = testing.shaped_arange((2, 3, 4), numpy, dtype)
        b = xp.asarray(a)
        # Make a computation here as just moving big-endian data back and forth
        # happens to work before the change in #5828
        return b + b

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_copy(self, xp, dtype, order):
        a = xp.zeros((2, 3, 4), dtype=dtype)
        b = xp.copy(a, order=order)
        a[1] = 1
        return b

    @testing.multi_gpu(2)
    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    def test_copy_multigpu(self, dtype, order):
        with cuda.Device(0):
            src = cupy.random.uniform(-1, 1, (2, 3)).astype(dtype)
        with cuda.Device(1):
            dst = cupy.copy(src, order)
        testing.assert_allclose(src, dst, rtol=0, atol=0)

    @testing.for_CF_orders()
    @testing.numpy_cupy_equal()
    def test_copy_order(self, xp, order):
        a = xp.zeros((2, 3, 4), order=order)
        b = xp.copy(a)
        return (b.flags.c_contiguous, b.flags.f_contiguous)

    @testing.numpy_cupy_array_equal()
    def test_asfortranarray_cuda_array_zero_dim(self, xp):
        a = xp.ones(())
        return xp.asfortranarray(a)

    @testing.for_all_dtypes_combination(['dtype_a', 'dtype_b'],
                                        no_complex=True)
    @testing.numpy_cupy_array_equal()
    def test_asfortranarray_cuda_array_zero_dim_dtype(
            self, xp, dtype_a, dtype_b):
        a = xp.ones((), dtype=dtype_a)
        return xp.asfortranarray(a, dtype=dtype_b)

    @testing.numpy_cupy_array_equal()
    def test_fromfile(self, xp):
        with tempfile.TemporaryFile() as fh:
            fh.write(b"\x00\x01\x02\x03\x04")
            fh.flush()
            fh.seek(0)
            return xp.fromfile(fh, dtype="u1")

    @testing.numpy_cupy_array_equal()
    def test_fromfunction(self, xp):
        def function(i, j): return i == j
        return xp.fromfunction(function, shape=(3, 3), dtype=int)

    @testing.numpy_cupy_array_equal()
    def test_fromiter(self, xp):
        iterable = (x*x for x in range(5))
        return xp.fromiter(iterable, float)

    @testing.numpy_cupy_array_equal()
    def test_fromstring(self, xp):
        return xp.fromstring('1 2', dtype=int, sep=' ')

    @testing.numpy_cupy_array_equal()
    def test_frombuffer(self, xp):
        return xp.frombuffer(b'\x01\x02', dtype=numpy.uint8)

    @testing.numpy_cupy_array_equal()
    def test_loadtxt(self, xp):
        with tempfile.TemporaryFile() as fh:
            fh.write(b"0 1\n2 3")
            fh.flush()
            fh.seek(0)
            return xp.loadtxt(fh, dtype="u1")

    @testing.numpy_cupy_array_equal()
    def test_genfromtxt(self, xp):
        with tempfile.TemporaryFile() as fh:
            fh.write(b"0 1\n2 3")
            fh.flush()
            fh.seek(0)
            return xp.genfromtxt(fh, dtype="u1")

    @testing.numpy_cupy_array_equal()
    def test_fromfile_big_endian(self, xp):
        with tempfile.TemporaryFile() as fh:
            fh.write(b"\x00\x00\x00\x01")
            fh.flush()
            fh.seek(0)
            a = xp.fromfile(fh, dtype='>u4')
            # Make a computation here as just moving big-endian data back and
            # forth happens to work before the change in #5828
            return a + a


max_cuda_array_interface_version = 3


@testing.parameterize(*testing.product({
    'ver': tuple(range(max_cuda_array_interface_version+1)),
    'strides': (False, None, True),
}))
@pytest.mark.skipif(
    cupy.cuda.runtime.is_hip, reason='HIP does not support this')
class TestCudaArrayInterface(unittest.TestCase):
    @testing.for_all_dtypes()
    def test_base(self, dtype):
        a = testing.shaped_arange((2, 3, 4), cupy, dtype)
        b = cupy.asarray(
            DummyObjectWithCudaArrayInterface(a, self.ver, self.strides))
        testing.assert_array_equal(a, b)

    @testing.for_all_dtypes()
    def test_not_copied(self, dtype):
        a = testing.shaped_arange((2, 3, 4), cupy, dtype)
        b = cupy.asarray(
            DummyObjectWithCudaArrayInterface(a, self.ver, self.strides))
        a.fill(0)
        testing.assert_array_equal(a, b)

    @testing.for_all_dtypes()
    def test_order(self, dtype):
        a = testing.shaped_arange((2, 3, 4), cupy, dtype)
        b = cupy.asarray(
            DummyObjectWithCudaArrayInterface(a, self.ver, self.strides),
            order='F')
        assert b.flags.f_contiguous
        testing.assert_array_equal(a, b)

    @testing.for_all_dtypes()
    def test_with_strides(self, dtype):
        a = testing.shaped_arange((2, 3, 4), cupy, dtype).T
        b = cupy.asarray(
            DummyObjectWithCudaArrayInterface(a, self.ver, self.strides))
        assert a.strides == b.strides
        assert a.nbytes == b.data.mem.size

    @testing.for_all_dtypes()
    def test_with_zero_size_array(self, dtype):
        a = testing.shaped_arange((0,), cupy, dtype)
        b = cupy.asarray(
            DummyObjectWithCudaArrayInterface(a, self.ver, self.strides))
        assert a.strides == b.strides
        assert a.nbytes == b.data.mem.size
        assert a.data.ptr == 0
        assert a.size == 0

    @testing.for_all_dtypes()
    def test_asnumpy(self, dtype):
        a = testing.shaped_arange((2, 3, 4), cupy, dtype)
        b = DummyObjectWithCudaArrayInterface(a, self.ver, self.strides)
        a_cpu = cupy.asnumpy(a)
        b_cpu = cupy.asnumpy(b)
        testing.assert_array_equal(a_cpu, b_cpu)

    def test_big_endian(self):
        a = cupy.array([0x1, 0x0, 0x0, 0x0], dtype=numpy.int8)
        dtype = numpy.dtype('>i4')
        shape = 1,
        strides = 4,
        data = a.data.ptr
        b = DummyObjectWithCudaArrayInterface(
            (shape, strides, dtype.str, dtype.descr, data),
            self.ver, self.strides)
        with pytest.raises(ValueError):
            cupy.asarray(b)


@testing.parameterize(*testing.product({
    'ver': tuple(range(1, max_cuda_array_interface_version+1)),
    'strides': (False, None, True),
}))
@pytest.mark.skipif(
    cupy.cuda.runtime.is_hip, reason='HIP does not support this')
class TestCudaArrayInterfaceMaskedArray(unittest.TestCase):
    # TODO(leofang): update this test when masked array is supported
    @testing.for_all_dtypes()
    def test_masked_array(self, dtype):
        a = testing.shaped_arange((2, 3, 4), cupy, dtype)
        mask = testing.shaped_arange((2, 3, 4), cupy, dtype)
        a = DummyObjectWithCudaArrayInterface(a, self.ver, self.strides, mask)
        with pytest.raises(ValueError) as ex:
            b = cupy.asarray(a)  # noqa
        assert 'does not support' in str(ex.value)


# marked slow as either numpy or cupy could go OOM in this test
@testing.slow
@pytest.mark.skipif(
    cupy.cuda.runtime.is_hip, reason='HIP does not support this')
class TestCudaArrayInterfaceBigArray(unittest.TestCase):
    def test_with_over_size_array(self):
        # real example from #3009
        size = 5 * 10**8
        a = testing.shaped_random((size,), cupy, cupy.float64)
        b = cupy.asarray(DummyObjectWithCudaArrayInterface(a, 2, None))
        testing.assert_array_equal(a, b)


@pytest.mark.skipif(
    cupy.cuda.runtime.is_hip, reason='HIP does not support this')
class DummyObjectWithCudaArrayInterface(object):
    def __init__(self, a, ver, include_strides=False, mask=None, stream=None):
        assert ver in tuple(range(max_cuda_array_interface_version+1))
        self.a = None
        if isinstance(a, cupy.ndarray):
            self.a = a
        else:
            self.shape, self.strides, self.typestr, self.descr, self.data = a
        self.ver = ver
        self.include_strides = include_strides
        self.mask = mask
        self.stream = stream

    @property
    def __cuda_array_interface__(self):
        if self.a is not None:
            desc = {
                'shape': self.a.shape,
                'typestr': self.a.dtype.str,
                'descr': self.a.dtype.descr,
                'data': (self.a.data.ptr, False),
                'version': self.ver,
            }
            if self.a.flags.c_contiguous:
                if self.include_strides is True:
                    desc['strides'] = self.a.strides
                elif self.include_strides is None:
                    desc['strides'] = None
                else:  # self.include_strides is False
                    pass
            else:  # F contiguous or neither
                desc['strides'] = self.a.strides
        else:
            desc = {
                'shape': self.shape,
                'typestr': self.typestr,
                'descr': self.descr,
                'data': (self.data, False),
                'version': self.ver,
            }
            if self.include_strides is True:
                desc['strides'] = self.strides
            elif self.include_strides is None:
                desc['strides'] = None
            else:  # self.include_strides is False
                pass
        if self.mask is not None:
            desc['mask'] = self.mask
        # The stream field is kept here for compliance. However, since the
        # synchronization is done via calling a cpdef function, which cannot
        # be mock-tested.
        if self.stream is not None:
            if self.stream is cuda.Stream.null:
                desc['stream'] = cuda.runtime.streamLegacy
            elif (not cuda.runtime.is_hip) and self.stream is cuda.Stream.ptds:
                desc['stream'] = cuda.runtime.streamPerThread
            else:
                desc['stream'] = self.stream.ptr
        return desc


@testing.parameterize(
    *testing.product({
        'ndmin': [0, 1, 2, 3],
        'copy': [True, False],
        'xp': [numpy, cupy]
    })
)
class TestArrayPreservationOfShape(unittest.TestCase):

    @testing.for_all_dtypes()
    def test_cupy_array(self, dtype):
        shape = 2, 3
        a = testing.shaped_arange(shape, self.xp, dtype)
        cupy.array(a, copy=self.copy, ndmin=self.ndmin)

        # Check if cupy.ndarray does not alter
        # the shape of the original array.
        assert a.shape == shape


@testing.parameterize(
    *testing.product({
        'ndmin': [0, 1, 2, 3],
        'copy': [True, False],
        'xp': [numpy, cupy]
    })
)
class TestArrayCopy(unittest.TestCase):

    @testing.for_all_dtypes()
    def test_cupy_array(self, dtype):
        a = testing.shaped_arange((2, 3), self.xp, dtype)
        actual = cupy.array(a, copy=self.copy, ndmin=self.ndmin)

        should_copy = (self.xp is numpy) or self.copy
        # TODO(Kenta Oono): Better determination of copy.
        is_copied = not ((actual is a) or (actual.base is a) or
                         (actual.base is a.base and a.base is not None))
        assert should_copy == is_copied


class TestArrayInvalidObject(unittest.TestCase):

    def test_invalid_type(self):
        a = numpy.array([1, 2, 3], dtype=object)
        with self.assertRaises(ValueError):
            cupy.array(a)



import contextlib
import os
import unittest

import numpy
import pytest

import cupy
from cupy import testing
import cupyx

try:
    import scipy.sparse
    scipy_available = True
except ImportError:
    scipy_available = False


class TestGetArrayModule(unittest.TestCase):

    def test_get_array_module_1(self):
        n1 = numpy.array([2], numpy.float32)
        c1 = cupy.array([2], numpy.float32)
        csr1 = cupyx.scipy.sparse.csr_matrix((5, 3), dtype=numpy.float32)

        assert numpy is cupy.get_array_module()
        assert numpy is cupy.get_array_module(n1)
        assert cupy is cupy.get_array_module(c1)
        assert cupy is cupy.get_array_module(csr1)

        assert numpy is cupy.get_array_module(n1, n1)
        assert cupy is cupy.get_array_module(c1, c1)
        assert cupy is cupy.get_array_module(csr1, csr1)

        assert cupy is cupy.get_array_module(n1, csr1)
        assert cupy is cupy.get_array_module(csr1, n1)
        assert cupy is cupy.get_array_module(c1, n1)
        assert cupy is cupy.get_array_module(n1, c1)
        assert cupy is cupy.get_array_module(c1, csr1)
        assert cupy is cupy.get_array_module(csr1, c1)

        if scipy_available:
            csrn1 = scipy.sparse.csr_matrix((5, 3), dtype=numpy.float32)

            assert numpy is cupy.get_array_module(csrn1)
            assert cupy is cupy.get_array_module(csrn1, csr1)
            assert cupy is cupy.get_array_module(csr1, csrn1)
            assert cupy is cupy.get_array_module(c1, csrn1)
            assert cupy is cupy.get_array_module(csrn1, c1)
            assert numpy is cupy.get_array_module(n1, csrn1)
            assert numpy is cupy.get_array_module(csrn1, n1)


class MockArray(numpy.lib.mixins.NDArrayOperatorsMixin):
    __array_priority__ = 20  # less than cupy.ndarray.__array_priority__

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        assert method == '__call__'
        name = ufunc.__name__
        return name, inputs, kwargs


class TestArrayUfunc:

    def test_add(self):
        x = cupy.array([3, 7])
        y = MockArray()
        assert x + y == ('add', (x, y), {})
        assert y + x == ('add', (y, x), {})
        y2 = y
        y2 += x
        assert y2 == ('add', (y, x), {'out': y})
        with pytest.raises(TypeError):
            x += y

    @pytest.mark.xfail(
        reason='cupy.ndarray.__array_ufunc__ does not support gufuncs yet')
    def test_matmul(self):
        x = cupy.array([3, 7])
        y = MockArray()
        assert x @ y == ('matmul', (x, y), {})
        assert y @ x == ('matmul', (y, x), {})
        y2 = y
        y2 @= x
        assert y2 == ('matmul', (y, x), {'out': y})
        with pytest.raises(TypeError):
            x @= y

    def test_lt(self):
        x = cupy.array([3, 7])
        y = MockArray()
        assert (x < y) == ('less', (x, y), {})
        assert (y < x) == ('less', (y, x), {})


class MockArray2:
    __array_ufunc__ = None

    def __add__(self, other):
        return 'add'

    def __radd__(self, other):
        return 'radd'

    def __matmul__(self, other):
        return 'matmul'

    def __rmatmul__(self, other):
        return 'rmatmul'

    def __lt__(self, other):
        return 'lt'

    def __gt__(self, other):
        return 'gt'


class TestArrayUfuncOptout:

    def test_add(self):
        x = cupy.array([3, 7])
        y = MockArray2()
        assert x + y == 'radd'
        assert y + x == 'add'

    def test_matmul(self):
        x = cupy.array([3, 7])
        y = MockArray2()
        assert x @ y == 'rmatmul'
        assert y @ x == 'matmul'

    def test_lt(self):
        x = cupy.array([3, 7])
        y = MockArray2()
        assert (x < y) == 'gt'
        assert (y < x) == 'lt'


class TestAsnumpy:

    def test_asnumpy(self):
        x = testing.shaped_random((2, 3, 4), cupy, cupy.float64)
        y = cupy.asnumpy(x)
        testing.assert_array_equal(x, y)

    def test_asnumpy_out(self):
        x = testing.shaped_random((2, 3, 4), cupy, cupy.float64)
        y = cupyx.empty_like_pinned(x)
        y = cupy.asnumpy(x, out=y)
        testing.assert_array_equal(x, y)
        assert isinstance(y.base, cupy.cuda.PinnedMemoryPointer)
        assert y.base.ptr == y.ctypes.data

    @pytest.mark.skipif(
        int(os.environ.get('CUPY_ENABLE_UMP', 0)) == 1,
        reason='blocking or not is irrelevant when zero-copy is on'
    )
    @pytest.mark.parametrize('blocking', (True, False))
    def test_asnumpy_blocking(self, blocking):
        prefactor = 4
        a = cupy.random.random(prefactor*128*1024*1024, dtype=cupy.float64)
        cupy.cuda.Device().synchronize()

        # Idea: perform D2H copy on a nonblocking stream, during which we try
        # to "corrupt" the host data via NumPy operation. If the copy is
        # properly ordered, corruption would not be possible. Here we craft a
        # problem size and use pinned memory to ensure the failure can be
        # always triggered. (The CUDART API reference ("API synchronization
        # behavior") states that copying between device and pageable memory
        # "might be" synchronous, whereas between device and page-locked
        # memory "should be" fully asynchronous.)
        s = cupy.cuda.Stream(non_blocking=True)
        with s:
            c = cupyx.empty_pinned(a.shape, dtype=a.dtype)
            cupy.asnumpy(a, out=c, blocking=blocking)
            c[c.size//2:] = -1.  # potential data race
        s.synchronize()

        a[c.size//2:] = -1.
        if not blocking:
            ctx = pytest.raises(AssertionError)
        else:
            ctx = contextlib.nullcontext()
        with ctx:
            assert cupy.allclose(a, c)



import sys as _sys

from cupy_backends.cuda.api cimport runtime
from cupy_backends.cuda._softlink cimport SoftLink


ctypedef int nvrtcResult
ctypedef void* nvrtcProgram
# TODO(kmaehashi): Remove this alias.
ctypedef nvrtcProgram Program

ctypedef const char* (*F_nvrtcGetErrorString)(nvrtcResult result) noexcept nogil  # NOQA
cdef F_nvrtcGetErrorString nvrtcGetErrorString

ctypedef nvrtcResult (*F_nvrtcVersion)(int *major, int *minor) noexcept nogil
cdef F_nvrtcVersion nvrtcVersion

ctypedef nvrtcResult (*F_nvrtcCreateProgram)(
    nvrtcProgram* prog, const char* src, const char* name, int numHeaders,
    const char** headers, const char** includeNames) noexcept nogil
cdef F_nvrtcCreateProgram nvrtcCreateProgram

ctypedef nvrtcResult (*F_nvrtcDestroyProgram)(nvrtcProgram *prog) noexcept nogil  # NOQA
cdef F_nvrtcDestroyProgram nvrtcDestroyProgram

ctypedef nvrtcResult (*F_nvrtcCompileProgram)(
    nvrtcProgram prog, int numOptions, const char** options) noexcept nogil
cdef F_nvrtcCompileProgram nvrtcCompileProgram

ctypedef nvrtcResult (*F_nvrtcGetPTXSize)(nvrtcProgram prog, size_t *ptxSizeRet) noexcept nogil  # NOQA
cdef F_nvrtcGetPTXSize nvrtcGetPTXSize

ctypedef nvrtcResult (*F_nvrtcGetPTX)(nvrtcProgram prog, char *ptx) noexcept nogil  # NOQA
cdef F_nvrtcGetPTX nvrtcGetPTX

ctypedef nvrtcResult (*F_nvrtcGetCUBINSize)(nvrtcProgram prog, size_t *cubinSizeRet) noexcept nogil  # NOQA
cdef F_nvrtcGetCUBINSize nvrtcGetCUBINSize

ctypedef nvrtcResult (*F_nvrtcGetCUBIN)(nvrtcProgram prog, char *cubin) noexcept nogil  # NOQA
cdef F_nvrtcGetCUBIN nvrtcGetCUBIN

ctypedef nvrtcResult (*F_nvrtcGetProgramLogSize)(nvrtcProgram prog, size_t* logSizeRet) noexcept nogil  # NOQA
cdef F_nvrtcGetProgramLogSize nvrtcGetProgramLogSize

ctypedef nvrtcResult (*F_nvrtcGetProgramLog)(nvrtcProgram prog, char* log) noexcept nogil  # NOQA
cdef F_nvrtcGetProgramLog nvrtcGetProgramLog

ctypedef nvrtcResult (*F_nvrtcAddNameExpression)(nvrtcProgram, const char*) noexcept nogil  # NOQA
cdef F_nvrtcAddNameExpression nvrtcAddNameExpression

ctypedef nvrtcResult (*F_nvrtcGetLoweredName)(nvrtcProgram, const char*, const char**) noexcept nogil  # NOQA
cdef F_nvrtcGetLoweredName nvrtcGetLoweredName

ctypedef nvrtcResult (*F_nvrtcGetNumSupportedArchs)(int* numArchs) noexcept nogil  # NOQA
cdef F_nvrtcGetNumSupportedArchs nvrtcGetNumSupportedArchs

ctypedef nvrtcResult (*F_nvrtcGetSupportedArchs)(int* supportedArchs) noexcept nogil  # NOQA
cdef F_nvrtcGetSupportedArchs nvrtcGetSupportedArchs

ctypedef nvrtcResult (*F_nvrtcGetNVVMSize)(nvrtcProgram prog, size_t *nvvmSizeRet) noexcept nogil  # NOQA
cdef F_nvrtcGetNVVMSize nvrtcGetNVVMSize

ctypedef nvrtcResult (*F_nvrtcGetNVVM)(nvrtcProgram prog, char *nvvm) noexcept nogil  # NOQA
cdef F_nvrtcGetNVVM nvrtcGetNVVM


cdef SoftLink _L = None
cdef inline void initialize() except *:
    global _L
    if _L is not None:
        return
    _L = _initialize()

cdef SoftLink _initialize():
    _L = _get_softlink()

    global nvrtcGetErrorString
    nvrtcGetErrorString = <F_nvrtcGetErrorString>_L.get('GetErrorString')
    global nvrtcVersion
    nvrtcVersion = <F_nvrtcVersion>_L.get('Version')
    global nvrtcCreateProgram
    nvrtcCreateProgram = <F_nvrtcCreateProgram>_L.get('CreateProgram')
    global nvrtcDestroyProgram
    nvrtcDestroyProgram = <F_nvrtcDestroyProgram>_L.get('DestroyProgram')
    global nvrtcCompileProgram
    nvrtcCompileProgram = <F_nvrtcCompileProgram>_L.get('CompileProgram')
    global nvrtcGetPTXSize
    nvrtcGetPTXSize = <F_nvrtcGetPTXSize>_L.get('GetPTXSize' if _L.prefix == 'nvrtc' else 'GetCodeSize')  # NOQA
    global nvrtcGetPTX
    nvrtcGetPTX = <F_nvrtcGetPTX>_L.get('GetPTX' if _L.prefix == 'nvrtc' else 'GetCode')  # NOQA
    global nvrtcGetCUBINSize
    nvrtcGetCUBINSize = <F_nvrtcGetCUBINSize>_L.get('GetCUBINSize')
    global nvrtcGetCUBIN
    nvrtcGetCUBIN = <F_nvrtcGetCUBIN>_L.get('GetCUBIN')
    global nvrtcGetProgramLogSize
    nvrtcGetProgramLogSize = <F_nvrtcGetProgramLogSize>_L.get('GetProgramLogSize')  # NOQA
    global nvrtcGetProgramLog
    nvrtcGetProgramLog = <F_nvrtcGetProgramLog>_L.get('GetProgramLog')
    global nvrtcAddNameExpression
    nvrtcAddNameExpression = <F_nvrtcAddNameExpression>_L.get('AddNameExpression')  # NOQA
    global nvrtcGetLoweredName
    nvrtcGetLoweredName = <F_nvrtcGetLoweredName>_L.get('GetLoweredName')
    global nvrtcGetNumSupportedArchs
    nvrtcGetNumSupportedArchs = <F_nvrtcGetNumSupportedArchs>_L.get('GetNumSupportedArchs')  # NOQA
    global nvrtcGetSupportedArchs
    nvrtcGetSupportedArchs = <F_nvrtcGetSupportedArchs>_L.get('GetSupportedArchs')  # NOQA
    global nvrtcGetNVVMSize
    nvrtcGetNVVMSize = <F_nvrtcGetNVVMSize>_L.get('GetNVVMSize')
    global nvrtcGetNVVM
    nvrtcGetNVVM = <F_nvrtcGetNVVM>_L.get('GetNVVM')

    return _L


cdef SoftLink _get_softlink():
    cdef int runtime_version
    cdef str prefix = 'nvrtc'
    cdef object libname = None

    if CUPY_CUDA_VERSION != 0:
        runtime_version = runtime._getCUDAMajorVersion()
        if runtime_version == 11:
            # CUDA 11.x (11.2+)
            if _sys.platform == 'linux':
                libname = 'libnvrtc.so.11.2'
            else:
                libname = 'nvrtc64_112_0.dll'
        elif runtime_version == 12:
            # CUDA 12.x
            if _sys.platform == 'linux':
                libname = 'libnvrtc.so.12'
            else:
                libname = 'nvrtc64_120_0.dll'
        elif runtime_version == 13:
            # CUDA 13.x
            if _sys.platform == 'linux':
                libname = 'libnvrtc.so.13'
            else:
                libname = 'nvrtc64_130_0.dll'
    elif CUPY_HIP_VERSION != 0:
        runtime_version = runtime.runtimeGetVersion()
        prefix = 'hiprtc'
        if runtime_version < 5_00_00000:
            # ROCm 4.x
            libname = 'libamdhip64.so.4'
        elif runtime_version < 6_00_00000:
            # ROCm 5.x
            libname = 'libamdhip64.so.5'
        elif runtime_version < 7_00_00000:
            # ROCm 6.x
            libname = 'libamdhip64.so.6'

    return SoftLink(libname, prefix, mandatory=True)



from math import sqrt, pi

import pytest
import numpy as np

import cupy
from cupy import testing
import cupyx.scipy.signal  # NOQA
import cupyx.scipy.signal as signal
from cupy.testing import assert_array_almost_equal, assert_allclose


try:
    import scipy
    import scipy.signal  # NOQA
except ImportError:
    pass

try:
    import mpmath  # NOQA
except ImportError:
    pass


@testing.with_requires("scipy")
class TestFindFreqs:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_docstring(self, xp, scp):
        ff = scp.signal.findfreqs
        return ff(xp.array([1, 0]), xp.array([1, 8, 25]), N=9)

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_ones(self, xp, scp):
        ff = scp.signal.findfreqs
        return ff(xp.array([1.0]), [1.0], N=8)


@testing.with_requires("scipy")
class TestFreqs:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp):
        w, h = scp.signal.freqs(xp.asarray([1.0]), xp.asarray([1.0]), worN=8)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_output(self, xp, scp):
        # 1st order low-pass filter: H(s) = 1 / (s + 1)
        w = xp.asarray([0.1, 1, 10, 100])
        num = xp.asarray([1])
        den = xp.asarray([1, 1])
        w, H = scp.signal.freqs(num, den, worN=w)
        return w, H

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_freq_range(self, xp, scp):
        # Test that freqresp() finds a reasonable frequency range.
        # 1st order low-pass filter: H(s) = 1 / (s + 1)
        # Expected range is from 0.01 to 10.
        num = xp.asarray([1])
        den = xp.asarray([1, 1])
        n = 10
        w, H = scp.signal.freqs(num, den, worN=n)
        return w, H

    @pytest.mark.parametrize('w', [8.0, 8.0 + 0j])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_w_or_N_types(self, xp, scp, w):
        w, h = scp.signal.freqs(xp.asarray([1.0]), xp.asarray([1.0]), worN=w)
        return w, h


@testing.with_requires("scipy")
class TestFreqs_zpk:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp):
        w, h = scp.signal.freqs_zpk(xp.asarray(
            [1.0]), xp.asarray([1.0]), xp.asarray([1.0]), worN=8)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_output(self, xp, scp):
        # 1st order low-pass filter: H(s) = 1 / (s + 1)
        w = xp.asarray([0.1, 1, 10, 100])
        z = xp.asarray([])
        p = xp.asarray([-1])
        k = 1
        w, H = scp.signal.freqs_zpk(z, p, k, worN=w)
        return w, H

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_freq_range(self, xp, scp):
        # Test that freqresp() finds a reasonable frequency range.
        # 1st order low-pass filter: H(s) = 1 / (s + 1)
        # Expected range is from 0.01 to 10.
        z = xp.asarray([])
        p = xp.asarray([-1])
        k = 1
        n = 10
        w, H = scp.signal.freqs_zpk(z, p, k, worN=n)
        return w, H

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_vs_freqs(self, xp, scp):
        z, p, k = scp.signal.cheby1(4, 5, 100, analog=True, output='zpk')
        w2, h2 = scp.signal.freqs_zpk(z, p, k)
        return w2, h2

    @pytest.mark.parametrize('w', [8.0, 8.0 + 0j])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_w_or_N_types(self, xp, scp, w):
        w, h = scp.signal.freqs_zpk(xp.asarray([]), xp.asarray([]), 1, worN=w)
        return w, h


@testing.with_requires("scipy")
class TestFreqz:

    def test_ticket1441(self):
        """Regression test for ticket 1441."""
        # Because freqz previously used arange instead of linspace,
        # when N was large, it would return one more point than
        # requested.
        N = 100000
        w, h = signal.freqz([1.0], worN=N)
        assert w.shape == (N,)

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp):
        w, h = scp.signal.freqz([1.0], worN=8)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic_2(self, xp, scp):
        w, h = scp.signal.freqz([1.0], worN=9)
        return w, h

    def test_basic_3(self):
        for a in [1, cupy.ones(2)]:
            w, h = signal.freqz(cupy.ones(2), a, worN=0)
            assert w.shape == (0,)
            assert h.shape == (0,)
            assert h.dtype == cupy.dtype('complex128')

    def test_basic_4(self):
        xp = cupy
        t = xp.linspace(0, 1, 4, endpoint=False)
        for b, a, h_whole in zip(
                ([1., 0, 0, 0], xp.sin(2 * xp.pi * t)),
                ([1., 0, 0, 0], [0.5, 0, 0, 0]),
                ([1., 1., 1., 1.], [0, -4j, 0, 4j])):
            w, h = signal.freqz(b, a, worN=4, whole=True)

            expected_w = xp.linspace(0, 2 * xp.pi, 4, endpoint=False)
            assert_array_almost_equal(w, expected_w)
            assert_array_almost_equal(h, h_whole)

            # simultaneously check int-like support
            w, h = signal.freqz(b, a, worN=xp.int32(4), whole=True)
            assert_array_almost_equal(w, expected_w)
            assert_array_almost_equal(h, h_whole)

            w, h = signal.freqz(b, a, worN=w, whole=True)
            assert_array_almost_equal(w, expected_w)
            assert_array_almost_equal(h, h_whole)

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic_whole(self, xp, scp):
        w, h = scp.signal.freqz([1.0], worN=8, whole=True)
        return w, h

    @pytest.mark.skip(reason='plot')
    def test_plot(self):

        def plot(w, h):
            assert_array_almost_equal(w, pi * cupy.arange(8.0) / 8)
            assert_array_almost_equal(h, cupy.ones(8))

        pytest.raises(ZeroDivisionError, signal.freqz, [1.0], worN=8,
                      plot=lambda w, h: 1 / 0)
        signal.freqz([1.0], worN=8, plot=plot)

    def test_fft_wrapping(self):
        # Some simple real FIR filters
        bs = list()  # filters
        as_ = list()
        hs_whole = list()
        hs_half = list()
        # 3 taps
        t = cupy.linspace(0, 1, 3, endpoint=False)
        bs.append(cupy.sin(2 * pi * t))
        as_.append(3.)
        hs_whole.append([0, -0.5j, 0.5j])
        hs_half.append([0, sqrt(1./12.), -0.5j])
        # 4 taps
        t = cupy.linspace(0, 1, 4, endpoint=False)
        bs.append(cupy.sin(2 * pi * t))
        as_.append(0.5)
        hs_whole.append([0, -4j, 0, 4j])
        hs_half.append([0, sqrt(8), -4j, -sqrt(8)])
        del t

        for ii, b in enumerate(bs):
            # whole
            a = as_[ii]
            expected_w = cupy.linspace(0, 2 * pi, len(b), endpoint=False)
            w, h = signal.freqz(b, a, worN=expected_w, whole=True)  # polyval
            err_msg = f'b = {b}, a={a}'
            assert_array_almost_equal(w, expected_w, err_msg=err_msg)
            assert_array_almost_equal(h, hs_whole[ii], err_msg=err_msg)

            w, h = signal.freqz(b, a, worN=len(b), whole=True)  # FFT
            assert_array_almost_equal(w, expected_w, err_msg=err_msg)
            assert_array_almost_equal(h, hs_whole[ii], err_msg=err_msg)

            # non-whole
            expected_w = cupy.linspace(0, pi, len(b), endpoint=False)
            w, h = signal.freqz(b, a, worN=expected_w, whole=False)  # polyval
            assert_array_almost_equal(w, expected_w, err_msg=err_msg)
            assert_array_almost_equal(h, hs_half[ii], err_msg=err_msg)

            w, h = signal.freqz(b, a, worN=len(b), whole=False)  # FFT
            assert_array_almost_equal(w, expected_w, err_msg=err_msg)
            assert_array_almost_equal(h, hs_half[ii], err_msg=err_msg)

    def test_fft_wrapping_2(self):
        # some random FIR filters (real + complex)
        # assume polyval is accurate
        rng = cupy.random.RandomState(0)
        for ii in range(2, 10):  # number of taps
            b = rng.randn(ii)
            for kk in range(2):
                a = rng.randn(1) if kk == 0 else rng.randn(3)
                for jj in range(2):
                    if jj == 1:
                        b = b + rng.randn(ii) * 1j
                    # whole
                    expected_w = cupy.linspace(0, 2 * pi, ii, endpoint=False)
                    w, expected_h = signal.freqz(
                        b, a, worN=expected_w, whole=True)
                    assert_array_almost_equal(w, expected_w)

                    w, h = signal.freqz(b, a, worN=ii, whole=True)
                    assert_array_almost_equal(w, expected_w)
                    assert_array_almost_equal(h, expected_h)

                    # half
                    expected_w = cupy.linspace(0, pi, ii, endpoint=False)
                    w, expected_h = signal.freqz(
                        b, a, worN=expected_w, whole=False)
                    assert_array_almost_equal(w, expected_w)

                    w, h = signal.freqz(b, a, worN=ii, whole=False)
                    assert_array_almost_equal(w, expected_w)
                    assert_array_almost_equal(h, expected_h)

    @pytest.mark.parametrize('whole', [True, False])
    @pytest.mark.parametrize('worN',
                             [16, 17, np.linspace(0, 1, 10), np.array([])])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_broadcasting1(self, xp, scp, whole, worN):
        # Test broadcasting with worN an integer or a 1-D array,
        # b and a are n-dimensional arrays.
        np.random.seed(123)
        b = np.random.rand(3, 5, 1)
        a = np.random.rand(2, 1)
        if xp == cupy:
            a = cupy.asarray(a)
            b = cupy.asarray(b)

            if isinstance(worN, np.ndarray):
                worN = cupy.asarray(worN)

        w, h = scp.signal.freqz(b, a, worN=worN, whole=whole)
        return w, h

    @pytest.mark.parametrize('whole', [True, False])
    @pytest.mark.parametrize('worN', [16, 17, np.linspace(0, 1, 10)])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_broadcasting2(self, xp, scp, whole, worN):
        # Test broadcasting with worN an integer or a 1-D array,
        # b is an n-dimensional array, and a is left at the default value.
        np.random.seed(123)
        b = np.random.rand(3, 5, 1)

        if xp == cupy:
            b = cupy.asarray(b)
            if isinstance(worN, np.ndarray):
                worN = cupy.asarray(worN)

        w, h = scp.signal.freqz(b, worN=worN, whole=whole)

        # with CuPy, division by a changes the strides:
        # fft_func(b, n=n_fft, axis=0)[:N] / a
        # fft_func(...)[:N] is F-ordered, but the ratio is C-ordered.
        # With Numpy, it remains F-ordered. Make a copy for the comparison.
        h = h.copy(order='F')

        return w, h

    @pytest.mark.parametrize('whole', [True, False])
    @pytest.mark.parametrize('worN', [16, np.linspace(0, 1, 16)])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_broadcasting3(self, xp, scp, whole, worN):
        # Test broadcasting where b.shape[-1] is the same length
        # as worN, and a is left at the default value.
        np.random.seed(123)
        N = 16
        b = np.random.rand(3, N)     # XXX: N is worN or len(worN) !

        if xp == cupy:
            b = cupy.asarray(b)

        w, h = scp.signal.freqz(b, worN=worN, whole=whole)
        assert w.size == N
        return w, h

    @pytest.mark.parametrize('whole', [True, False])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_broadcasting4(self, xp, scp, whole):
        # Test broadcasting with worN a 2-D array.
        np.random.seed(123)
        b = np.random.rand(4, 2, 1, 1)
        a = np.random.rand(5, 2, 1, 1)

        wh = []

        for worN in [np.random.rand(6, 7), np.empty((6, 0))]:
            if xp == cupy:
                a = cupy.asarray(a)
                b = cupy.asarray(b)
                worN = cupy.asarray(worN)

            w, h = scp.signal.freqz(b, a, worN=worN, whole=whole)
            wh.append(w)
            wh.append(h)
        return wh

    def test_backward_compat(self):
        # For backward compatibility, test if None act as a wrapper for default
        w1, h1 = signal.freqz([1.0], 1)
        w2, h2 = signal.freqz([1.0], 1, None)
        assert_array_almost_equal(w1, w2)
        assert_array_almost_equal(h1, h2)

    def test_fs_param(self):
        fs = 900
        b = [0.039479155677484369, 0.11843746703245311, 0.11843746703245311,
             0.039479155677484369]
        a = [1.0, -1.3199152021838287, 0.80341991081938424,
             -0.16767146321568049]

        # N = None, whole=False
        w1, h1 = signal.freqz(b, a, fs=fs)
        w2, h2 = signal.freqz(b, a)
        testing.assert_allclose(h1, h2)
        testing.assert_allclose(
            w1, cupy.linspace(0, fs/2, 512, endpoint=False))

        # N = None, whole=True
        w1, h1 = signal.freqz(b, a, whole=True, fs=fs)
        w2, h2 = signal.freqz(b, a, whole=True)
        testing.assert_allclose(h1, h2)
        testing.assert_allclose(w1, cupy.linspace(0, fs, 512, endpoint=False))

        # N = 5, whole=False
        w1, h1 = signal.freqz(b, a, 5, fs=fs)
        w2, h2 = signal.freqz(b, a, 5)
        testing.assert_allclose(h1, h2)
        testing.assert_allclose(w1, cupy.linspace(0, fs/2, 5, endpoint=False))

        # N = 5, whole=True
        w1, h1 = signal.freqz(b, a, 5, whole=True, fs=fs)
        w2, h2 = signal.freqz(b, a, 5, whole=True)
        testing.assert_allclose(h1, h2)
        testing.assert_allclose(w1, cupy.linspace(0, fs, 5, endpoint=False))

        # w is an array_like
        for w in ([123], (123,), cupy.array([123]), (50, 123, 230),
                  cupy.array([50, 123, 230])):
            w1, h1 = signal.freqz(b, a, w, fs=fs)
            w2, h2 = signal.freqz(b, a, 2 * pi * cupy.array(w) / fs)
            testing.assert_allclose(h1, h2)
            testing.assert_allclose(w, w1)

    @pytest.mark.parametrize('N',
                             [7, cupy.int8(7), cupy.int16(7), cupy.int32(7),
                              cupy.int64(7), cupy.array(7),
                              8, cupy.int8(8), cupy.int16(8), cupy.int32(8),
                              cupy.int64(8), cupy.array(8)
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_w_or_N_types(self, xp, scp, N):
        # Measure at 7 (polyval) or 8 (fft) equally-spaced points

        if xp == np and isinstance(N, cupy.ndarray):
            N = N.get()

        w, h = scp.signal.freqz([1.0], worN=N)
        w1, h1 = scp.signal.freqz([1.0], worN=N, fs=100)

        return w, h, w1, h1

    @pytest.mark.parametrize('N',
                             [7, cupy.int8(7), cupy.int16(7), cupy.int32(7),
                              cupy.int64(7), cupy.array(7),
                              8, cupy.int8(8), cupy.int16(8), cupy.int32(8),
                              cupy.int64(8), cupy.array(8)
                              ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_w_or_N_types_2(self, xp, scp, N):
        if xp == np and isinstance(N, cupy.ndarray):
            N = N.get()

        w, h = scp.signal.freqz([1.0], worN=N)
        return w, h

    @pytest.mark.parametrize("w", [8.0, 8.0 + 0j])
    @testing.with_requires('scipy<1.16')
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_w_or_N_types_3(self, xp, scp, w):
        # Measure at frequency 8 Hz
        # Only makes sense when fs is specified
        w, h = scp.signal.freqz([1.0], worN=w, fs=100)
        return w, h

    @pytest.mark.parametrize("worN", [8, 9])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_nyquist(self, xp, scp, worN):
        w, h = scp.signal.freqz([1.0], worN=worN, include_nyquist=True)
        return w, h

    @pytest.mark.parametrize("nyq", [True, False])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_nyquist_1(self, xp, scp, nyq):
        w, h = scp.signal.freqz([1.0], worN=8, whole=True, include_nyquist=nyq)
        return w, h

    @pytest.mark.parametrize("a", [1, cupy.ones(2)])
    def test_nyquist_2(self, a):
        w, h = signal.freqz(cupy.ones(2), a, worN=0, include_nyquist=True)
        assert w.shape == (0,)
        assert h.shape == (0,)
        assert h.dtype == cupy.dtype('complex128')


@testing.with_requires("scipy")
class TestFreqz_zpk:

    def test_ticket1441(self):
        """Regression test for ticket 1441."""
        # Because freqz previously used arange instead of linspace,
        # when N was large, it would return one more point than
        # requested.
        N = 100000
        w, h = signal.freqz_zpk([0.5], [0.5], 1.0, worN=N)
        assert w.shape == (N,)

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp):
        w, h = scp.signal.freqz_zpk([0.5], [0.5], 1.0, worN=8)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic_whole(self, xp, scp):
        w, h = scp.signal.freqz_zpk([0.5], [0.5], 1.0, worN=8, whole=True)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_vs_freqz(self, xp, scp):
        z, p, k = scp.signal.cheby1(4, 5, 0.5, analog=False, output='zpk')
        w, h = scp.signal.freqz_zpk(z, p, k)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_vs_freqz_zpk(self, xp, scp):
        z, p, k = scp.signal.cheby1(4, 5, 0.5, analog=False, output='zpk')
        w2, h2 = scp.signal.freqz_zpk(z, p, k)
        return w2, h2

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_backward_compat(self, xp, scp):
        # For backward compatibility, test if None act as a wrapper for default
        w1, h1 = scp.signal.freqz_zpk([0.5], [0.5], 1.0)
        w2, h2 = scp.signal.freqz_zpk([0.5], [0.5], 1.0, None)
        return w1, h1, w2, h2

    def test_fs_param(self):
        fs = 900
        z = [-1, -1, -1]
        p = [0.4747869998473389+0.4752230717749344j, 0.37256600288916636,
             0.4747869998473389-0.4752230717749344j]
        k = 0.03934683014103762

        # N = None, whole=False
        w1, h1 = signal.freqz_zpk(z, p, k, whole=False, fs=fs)
        w2, h2 = signal.freqz_zpk(z, p, k, whole=False)
        assert_allclose(h1, h2)
        assert_allclose(w1, cupy.linspace(0, fs/2, 512, endpoint=False))

        # N = None, whole=True
        w1, h1 = signal.freqz_zpk(z, p, k, whole=True, fs=fs)
        w2, h2 = signal.freqz_zpk(z, p, k, whole=True)
        assert_allclose(h1, h2)
        assert_allclose(w1, cupy.linspace(0, fs, 512, endpoint=False))

        # N = 5, whole=False
        w1, h1 = signal.freqz_zpk(z, p, k, 5, fs=fs)
        w2, h2 = signal.freqz_zpk(z, p, k, 5)
        assert_allclose(h1, h2)
        assert_allclose(w1, cupy.linspace(0, fs/2, 5, endpoint=False))

        # N = 5, whole=True
        w1, h1 = signal.freqz_zpk(z, p, k, 5, whole=True, fs=fs)
        w2, h2 = signal.freqz_zpk(z, p, k, 5, whole=True)
        assert_allclose(h1, h2)
        assert_allclose(w1, cupy.linspace(0, fs, 5, endpoint=False))

        # w is an array_like
        for w in ([123], (123,), cupy.array([123]), (50, 123, 230),
                  cupy.array([50, 123, 230])):
            w1, h1 = signal.freqz_zpk(z, p, k, w, fs=fs)
            w2, h2 = signal.freqz_zpk(z, p, k, 2*pi*cupy.array(w)/fs)
            assert_allclose(h1, h2)
            assert_allclose(w, w1)

    def test_w_or_N_types(self):
        # Measure at 8 equally-spaced points
        for N in (8, cupy.int8(8), cupy.int16(8), cupy.int32(8), cupy.int64(8),
                  cupy.array(8)):

            w, h = signal.freqz_zpk([], [], 1, worN=N)
            assert_array_almost_equal(w, pi * cupy.arange(8) / 8.)
            assert_array_almost_equal(h, cupy.ones(8))

            w, h = signal.freqz_zpk([], [], 1, worN=N, fs=100)
            assert_array_almost_equal(
                w, cupy.linspace(0, 50, 8, endpoint=False))
            assert_array_almost_equal(h, cupy.ones(8))

        # Measure at frequency 8 Hz
        for w in (8.0, 8.0+0j):
            # Only makes sense when fs is specified
            w_out, h = signal.freqz_zpk([], [], 1, worN=w, fs=100)
            assert_array_almost_equal(w_out, [8])
            assert_array_almost_equal(h, [1])


@testing.with_requires("scipy>=1.8")
class TestSOSFreqz:

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_sosfreqz_basic(self, xp, scp):
        # Compare the results of freqz and sosfreqz for a low order
        # Butterworth filter.
        N = 500
        sos = scp.signal.butter(4, 0.2, output='sos')
        w2, h2 = scp.signal.sosfreqz(sos, worN=N)
        return w2, h2

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_sosfreqz_basic_1(self, xp, scp):
        N = 500
        sos = scp.signal.ellip(3, 1, 30, (0.2, 0.3),
                               btype='bandpass', output='sos')
        w2, h2 = scp.signal.sosfreqz(sos, worN=N)
        return w2, h2

    # Compare sosfreqz output against expected values for different
    # filter types

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_sosfrez_design_cheb2(self, xp, scp):
        N, Wn = scp.signal.cheb2ord([0.1, 0.6], [0.2, 0.5], 3, 60)
        sos = scp.signal.cheby2(N, 60, Wn, 'stop', output='sos')
        w, h = scp.signal.sosfreqz(sos)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_sosfrez_design_cheb2_2(self, xp, scp):
        N, Wn = scp.signal.cheb2ord([0.1, 0.6], [0.2, 0.5], 3, 150)
        sos = scp.signal.cheby2(N, 150, Wn, 'stop', output='sos')
        w, h = scp.signal.sosfreqz(sos)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_sosfrez_design_cheb1(self, xp, scp):
        N, Wn = scp.signal.cheb1ord(0.2, 0.3, 3, 40)
        sos = scp.signal.cheby1(N, 3, Wn, 'low', output='sos')
        w, h = scp.signal.sosfreqz(sos)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_sosfrez_design_cheb1_2(self, xp, scp):
        N, Wn = scp.signal.cheb1ord(0.2, 0.3, 1, 150)
        sos = scp.signal.cheby1(N, 1, Wn, 'low', output='sos')
        w, h = scp.signal.sosfreqz(sos)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_sosfrez_design_butter(self, xp, scp):
        # adapted from buttord
        N, Wn = scp.signal.buttord([0.2, 0.5], [0.14, 0.6], 3, 40)
        sos = scp.signal.butter(N, Wn, 'band', output='sos')
        w, h = scp.signal.sosfreqz(sos)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_sosfrez_design_butter_2(self, xp, scp):
        N, Wn = scp.signal.buttord([0.2, 0.5], [0.14, 0.6], 3, 100)
        sos = scp.signal.butter(N, Wn, 'band', output='sos')
        w, h = scp.signal.sosfreqz(sos)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_sosfreqz_design_ellip(self, xp, scp):
        N, Wn = scp.signal.ellipord(0.3, 0.1, 3, 60)
        sos = scp.signal.ellip(N, 0.3, 60, Wn, 'high', output='sos')
        w, h = scp.signal.sosfreqz(sos)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_sosfreqz_design_ellip_2(self, xp, scp):
        N, Wn = scp.signal.ellipord(0.3, 0.2, 3, 60)
        sos = scp.signal.ellip(N, 0.3, 60, Wn, 'high', output='sos')
        w, h = scp.signal.sosfreqz(sos)
        return w, h

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_sosfreqz_design_ellip_3(self, xp, scp):
        N, Wn = scp.signal.ellipord(0.3, 0.2, .5, 150)
        sos = scp.signal.ellip(N, .5, 150, Wn, 'high', output='sos')
        w, h = scp.signal.sosfreqz(sos)
        return w, h

    @testing.with_requires("mpmath > 0.10")
    def test_sos_freqz_against_mp(self):
        # Compare the result of sosfreqz applied to a high order Butterworth
        # filter against the result computed using mpmath.  (signal.freqz fails
        # miserably with such high order filters.)
        from . import mpsig
        N = 500
        order = 25
        Wn = 0.15
        with mpmath.workdps(80):
            z_mp, p_mp, k_mp = mpsig.butter_lp(order, Wn)
            w_mp, h_mp = mpsig.zpkfreqz(z_mp, p_mp, k_mp, N)
        w_mp = np.array([float(x) for x in w_mp])
        h_mp = np.array([complex(x) for x in h_mp])

        sos = cupyx.scipy.signal.butter(order, Wn, output='sos')
        w, h = cupyx.scipy.signal.sosfreqz(sos, worN=N)
        assert_allclose(w, w_mp, rtol=1e-12, atol=1e-14)
        assert_allclose(h, h_mp, rtol=1e-12, atol=1e-14)

    def _get_fs_sos(self):
        fs = 900
        sos = [[0.03934683014103762, 0.07869366028207524, 0.03934683014103762,
                1.0, -0.37256600288916636, 0.0],
               [1.0, 1.0, 0.0, 1.0, -0.9495739996946778, 0.45125966317124144]]
        return fs, sos

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_fs_param(self, xp, scp):
        fs, sos = self._get_fs_sos()

        # N = None, whole=False
        w1, h1 = scp.signal.sosfreqz(sos, fs=fs)
        w2, h2 = scp.signal.sosfreqz(sos)
        return w1, h1, w2, h2

    @testing.numpy_cupy_allclose(scipy_name="scp", atol=1e-10)
    def test_fs_param_2(self, xp, scp):
        fs, sos = self._get_fs_sos()

        # N = None, whole=True
        w1, h1 = scp.signal.sosfreqz(sos, whole=True, fs=fs)
        w2, h2 = scp.signal.sosfreqz(sos, whole=True)
        return w1, h1, w2, h2

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_fs_param_3(self, xp, scp):
        fs, sos = self._get_fs_sos()
        # N = 5, whole=False
        w1, h1 = scp.signal.sosfreqz(sos, 5, fs=fs)
        w2, h2 = scp.signal.sosfreqz(sos, 5)
        return w1, h1, w2, h2

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_fs_param_4(self, xp, scp):
        fs, sos = self._get_fs_sos()
        # N = 5, whole=True
        w1, h1 = scp.signal.sosfreqz(sos, 5, whole=True, fs=fs)
        w2, h2 = scp.signal.sosfreqz(sos, 5, whole=True)
        return w1, h1, w2, h2

    @pytest.mark.parametrize("w", [[123], (123,),
                                   np.array([123]), (50, 123, 230),
                                   np.array([50, 123, 230])])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_fs_param_5(self, xp, scp, w):
        fs, sos = self._get_fs_sos()

        if xp == cupy and isinstance(w, np.ndarray):
            w = cupy.asarray(w)

        # w is an array_like
        w1, h1 = scp.signal.sosfreqz(sos, w, fs=fs)
        w2, h2 = scp.signal.sosfreqz(sos, 2*pi*xp.array(w)/fs)
        return w1, h1, w2, h2

    def test_w_or_N_types(self):
        # Measure at 7 (polyval) or 8 (fft) equally-spaced points
        for N in (7, cupy.int8(7), cupy.int16(7), cupy.int32(7), cupy.int64(7),
                  #     cupy.array(7),
                  8, cupy.int8(8), cupy.int16(8), cupy.int32(8), cupy.int64(8),
                  #     cupy.array(8)
                  ):

            w, h = signal.sosfreqz([1, 0, 0, 1, 0, 0], worN=N)
            assert_array_almost_equal(w, pi * cupy.arange(N) / N)
            assert_array_almost_equal(h, cupy.ones(N))

            w, h = signal.sosfreqz([1, 0, 0, 1, 0, 0], worN=N, fs=100)
            assert_array_almost_equal(
                w, cupy.linspace(0, 50, N, endpoint=False))
            assert_array_almost_equal(h, cupy.ones(N))

        # Measure at frequency 8 Hz
        for w in (8.0, 8.0+0j):
            # Only makes sense when fs is specified
            w_out, h = signal.sosfreqz([1, 0, 0, 1, 0, 0], worN=w, fs=100)
            assert_array_almost_equal(w_out, [8])
            assert_array_almost_equal(h, [1])


@testing.with_requires('scipy')
class TestGroupDelay:
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_identity_filter(self, xp, scp):
        w1, gd1 = scp.signal.group_delay((1, 1))
        w2, gd2 = scp.signal.group_delay((1, 1), whole=True)
        return w1, gd1, w2, gd2

    @pytest.mark.skip(reason='firwin is not available on CuPy')
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_fir(self, xp, scp):
        # Let's design linear phase FIR and check that the group delay
        # is constant.
        N = 100
        b = scp.signal.firwin(N + 1, 0.1)
        w, gd = scp.signal.group_delay((b, 1))
        return w, gd

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_iir(self, xp, scp):
        # Let's design Butterworth filter and test the group delay at
        # some points against MATLAB answer.
        b, a = scp.signal.butter(4, 0.1)
        w = xp.linspace(0, xp.pi, num=10, endpoint=False)
        w, gd = scp.signal.group_delay((b, a), w=w)
        return w, gd

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_backward_compat(self, xp, scp):
        # For backward compatibility, test if None act as a wrapper for default
        w1, gd1 = scp.signal.group_delay((1, 1))
        w2, gd2 = scp.signal.group_delay((1, 1), None)
        return w1, gd1, w2, gd2

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_fs_param(self, xp, scp):
        # Let's design Butterworth filter and test the group delay at
        # some points against the normalized frequency answer.
        b, a = scp.signal.butter(4, 4800, fs=96000)
        w = xp.linspace(0, 96000 / 2, num=10, endpoint=False)
        w, gd = scp.signal.group_delay((b, a), w=w, fs=96000)
        return w, gd

    @pytest.mark.parametrize(
        'type_', [None, 'int8', 'int16', 'int32', 'int64'])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_N_types(self, type_, xp, scp):
        # Measure at 8 equally-spaced points
        N = 8
        if type_ is not None:
            wrapper = getattr(xp, type_)
            N = wrapper(N)

        w, gd = scp.signal.group_delay((1, 1), N)
        return w, gd

    @pytest.mark.parametrize('w', [8.0, 8.0+0j])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    @testing.with_requires('scipy>=1.8')
    def test_w_types(self, w, xp, scp):
        # Measure at frequency 8 rad/sec
        w_out, gd = scp.signal.group_delay((1, 1), w)
        return w_out, gd


@testing.with_requires('scipy')
class TestGammatone:
    # Test erroneous input cases.
    def test_invalid_input(self):
        for scp in [cupyx.scipy, scipy]:
            # Cutoff frequency is <= 0 or >= fs / 2.
            fs = 16000
            for args in [
                    (-fs, 'iir'), (0, 'fir'), (fs / 2, 'iir'), (fs, 'fir')]:
                with pytest.raises(ValueError, match='The frequency must be '
                                   'between '):
                    scp.signal.gammatone(*args, fs=fs)

            # Filter type is not fir or iir
            for args in [(440, 'fie'), (220, 'it')]:
                with pytest.raises(ValueError, match='ftype must be '):
                    scp.signal.gammatone(*args, fs=fs)

            # Order is <= 0 or > 24 for FIR filter.
            for args in [(440, 'fir', -50), (220, 'fir', 0), (110, 'fir', 25),
                         (55, 'fir', 50)]:
                with pytest.raises(ValueError, match='Invalid order: '):
                    scp.signal.gammatone(*args, numtaps=None, fs=fs)

    # Verify that the filter's frequency response is approximately
    # 1 at the cutoff frequency.
    @pytest.mark.parametrize('ftype', ['fir', 'iir'])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_frequency_response(self, ftype, xp, scp):
        fs = 16000
        # Create a gammatone filter centered at 1000 Hz.
        b, a = scp.signal.gammatone(1000, ftype, fs=fs)
        return b, xp.asarray(a)

    # All built-in IIR filters are real, so should have perfectly
    # symmetrical poles and zeros. Then ba representation (using
    # numpy.poly) will be purely real instead of having negligible
    # imaginary parts.
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_iir_symmetry(self, xp, scp):
        b, a = scp.signal.gammatone(440, 'iir', fs=24000)
        return b, a

    # Verify FIR filter coefficients with the paper's
    # Mathematica implementation
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_fir_ba_output(self, xp, scp):
        b, _ = scp.signal.gammatone(15, 'fir', fs=1000)
        return b

    # Verify IIR filter coefficients with the paper's MATLAB implementation
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_iir_ba_output(self, xp, scp):
        b, a = scp.signal.gammatone(440, 'iir', fs=16000)
        return b, a



import pytest

from cupy import array_api as xp


@pytest.mark.parametrize(
    "obj, ind, axis, expected",
    [
        ([0, 1, 2, 3], [1, 3], -1, [1, 3]),
        ([0, 1, 2, 3], [2, 0], 0, [2, 0]),
        ([[0, 1, 2, 3]], [0, 0, 0], 0, [[0, 1, 2, 3]] * 3),
        ([[0, 1, 2, 3]], [1, 2, 1], 1, [[1, 2, 1]]),
    ],
)
def test_take(obj, ind, axis, expected):
    """
    Tests xp.take function
    """
    x = xp.asarray(obj)
    ind = xp.asarray(ind)
    out = xp.take(x, ind, axis=axis)
    assert xp.all(out == xp.asarray(expected))



from cupyx.scipy.signal.windows._windows import general_cosine  # NOQA
from cupyx.scipy.signal.windows._windows import boxcar  # NOQA
from cupyx.scipy.signal.windows._windows import triang  # NOQA
from cupyx.scipy.signal.windows._windows import parzen  # NOQA
from cupyx.scipy.signal.windows._windows import bohman  # NOQA
from cupyx.scipy.signal.windows._windows import blackman  # NOQA
from cupyx.scipy.signal.windows._windows import nuttall  # NOQA
from cupyx.scipy.signal.windows._windows import blackmanharris  # NOQA
from cupyx.scipy.signal.windows._windows import flattop  # NOQA
from cupyx.scipy.signal.windows._windows import bartlett  # NOQA
from cupyx.scipy.signal.windows._windows import hann  # NOQA
from cupyx.scipy.signal.windows._windows import tukey  # NOQA
from cupyx.scipy.signal.windows._windows import barthann  # NOQA
from cupyx.scipy.signal.windows._windows import general_hamming  # NOQA
from cupyx.scipy.signal.windows._windows import hamming  # NOQA
from cupyx.scipy.signal.windows._windows import kaiser  # NOQA
from cupyx.scipy.signal.windows._windows import gaussian  # NOQA
from cupyx.scipy.signal.windows._windows import general_gaussian  # NOQA
from cupyx.scipy.signal.windows._windows import chebwin  # NOQA
from cupyx.scipy.signal.windows._windows import cosine  # NOQA
from cupyx.scipy.signal.windows._windows import exponential  # NOQA
from cupyx.scipy.signal.windows._windows import taylor  # NOQA
from cupyx.scipy.signal.windows._windows import get_window  # NOQA



# mypy: ignore-errors

import contextlib
import logging
import os
import platform
import re
import shlex
import shutil
import subprocess
import sys
import tempfile
from typing import List, Set

import cupy_builder
import cupy_builder.install_utils as utils
from cupy_builder import _environment
from cupy_builder._context import Context


if os.environ.get('CONDA_BUILD', '0') == '1':
    logging.basicConfig(level=logging.DEBUG)

PLATFORM_LINUX = sys.platform.startswith('linux')
PLATFORM_WIN32 = sys.platform.startswith('win32')

minimum_cudnn_version = 7600
maximum_cudnn_version = 8970

minimum_hip_version = 305  # for ROCm 3.5.0+

_cuda_path = 'NOT_INITIALIZED'
_rocm_path = 'NOT_INITIALIZED'
_compiler_base_options = None


# Using tempfile.TemporaryDirectory would cause an error during cleanup
# due to a bug: https://bugs.python.org/issue26660
@contextlib.contextmanager
def _tempdir():
    temp_dir = tempfile.mkdtemp()
    try:
        yield temp_dir
    finally:
        shutil.rmtree(temp_dir, ignore_errors=True)


def get_rocm_path():
    global _rocm_path

    # Use a magic word to represent the cache not filled because None is a
    # valid return value.
    if _rocm_path != 'NOT_INITIALIZED':
        return _rocm_path

    _rocm_path = os.environ.get('ROCM_HOME', '')
    return _rocm_path


def get_cuda_path():
    global _cuda_path

    # Use a magic word to represent the cache not filled because None is a
    # valid return value.
    if _cuda_path != 'NOT_INITIALIZED':
        return _cuda_path

    nvcc_path = utils.search_on_path(('nvcc', 'nvcc.exe'))
    cuda_path_default = None
    if nvcc_path is not None:
        cuda_path_default = os.path.normpath(
            os.path.join(os.path.dirname(nvcc_path), '..'))

    cuda_path = os.environ.get('CUDA_PATH', '')  # Nvidia default on Windows
    if len(cuda_path) > 0 and cuda_path != cuda_path_default:
        utils.print_warning(
            'nvcc path != CUDA_PATH',
            'nvcc path: %s' % cuda_path_default,
            'CUDA_PATH: %s' % cuda_path)

    if os.path.exists(cuda_path):
        _cuda_path = cuda_path
    elif cuda_path_default is not None:
        _cuda_path = cuda_path_default
    elif os.path.exists('/usr/local/cuda'):
        _cuda_path = '/usr/local/cuda'
    else:
        _cuda_path = None

    return _cuda_path


def get_nvcc_path() -> List[str]:
    nvcc = os.environ.get('NVCC', None)
    if nvcc:
        return shlex.split(nvcc)

    cuda_path = get_cuda_path()
    if cuda_path is None:
        return None

    if PLATFORM_WIN32:
        nvcc_bin = 'bin/nvcc.exe'
    else:
        nvcc_bin = 'bin/nvcc'

    nvcc_path = os.path.join(cuda_path, nvcc_bin)
    if os.path.exists(nvcc_path):
        return [nvcc_path]
    else:
        return None


def get_hipcc_path() -> List[str]:
    hipcc = os.environ.get('HIPCC', None)
    if hipcc:
        return shlex.split(hipcc)

    rocm_path = get_rocm_path()
    if rocm_path is None:
        return None

    if PLATFORM_WIN32:
        hipcc_bin = 'bin/hipcc.exe'
    else:
        hipcc_bin = 'bin/hipcc'

    hipcc_path = os.path.join(rocm_path, hipcc_bin)
    if os.path.exists(hipcc_path):
        return [hipcc_path]
    else:
        return None


def get_compiler_setting(ctx: Context, use_hip):
    cuda_path = None
    rocm_path = None

    if use_hip:
        rocm_path = get_rocm_path()
    else:
        cuda_path = get_cuda_path()

    include_dirs = ctx.include_dirs.copy()
    library_dirs = ctx.library_dirs.copy()
    define_macros = []
    extra_compile_args = []

    if cuda_path:
        include_dirs.append(os.path.join(cuda_path, 'include'))
        if PLATFORM_WIN32:
            library_dirs.append(os.path.join(cuda_path, 'bin'))
            library_dirs.append(os.path.join(cuda_path, 'lib', 'x64'))
        else:
            library_dirs.append(os.path.join(cuda_path, 'lib64'))
            library_dirs.append(os.path.join(cuda_path, 'lib'))

    if rocm_path:
        include_dirs.append(os.path.join(rocm_path, 'include'))
        include_dirs.append(os.path.join(rocm_path, 'include', 'hip'))
        include_dirs.append(os.path.join(rocm_path, 'include', 'rocrand'))
        include_dirs.append(os.path.join(rocm_path, 'include', 'hiprand'))
        include_dirs.append(os.path.join(rocm_path, 'include', 'roctracer'))
        include_dirs.append(os.path.join(rocm_path, 'include', 'hipblas'))
        include_dirs.append(os.path.join(rocm_path, 'include', 'hipsparse'))
        include_dirs.append(os.path.join(rocm_path, 'include', 'hipfft'))
        include_dirs.append(os.path.join(rocm_path, 'include', 'rocsolver'))
        include_dirs.append(os.path.join(rocm_path, 'include', 'rccl'))
        library_dirs.append(os.path.join(rocm_path, 'lib'))

    if use_hip:
        # ROCm 5.3 and above requires c++14
        extra_compile_args.append('-std=c++14')

    if PLATFORM_WIN32:
        nvtx_path = _environment.get_nvtx_path()
        if nvtx_path is not None and os.path.exists(nvtx_path):
            include_dirs.append(os.path.join(nvtx_path, 'include'))
        else:
            define_macros.append(('CUPY_NO_NVTX', '1'))

    # For CUB, we need the complex and CUB headers. The search precedence for
    # the latter is:
    #   - for CUDA: CuPy's CUB (and Thrust) bundle
    #   - for ROCm: built-in CUB
    # Note that starting CuPy v8 we no longer use CUB_PATH, and starting v13
    # we no longer use Thrust/CUB bundled in CUDA.

    # for <cupy/complex.cuh>
    cupy_header = os.path.join(
        cupy_builder.get_context().source_root, 'cupy/_core/include')
    global _jitify_path
    _jitify_path = os.path.join(cupy_header, 'cupy/_jitify')
    global _cub_path
    if rocm_path:
        _cub_path = os.path.join(rocm_path, 'include', 'hipcub')
        if not os.path.exists(_cub_path):
            raise Exception('Please install hipCUB and retry')
        _thrust_path = None
        _libcudacxx_path = None
    else:
        # all bundled together under cccl
        _cub_path = os.path.join(cupy_header, 'cupy/_cccl/cub')
        _thrust_path = os.path.join(cupy_header, 'cupy/_cccl/thrust')
        _libcudacxx_path = os.path.join(cupy_header, 'cupy/_cccl/libcudacxx')
    include_dirs.insert(0, cupy_header)
    include_dirs.insert(0, _cub_path)
    if _thrust_path and _libcudacxx_path:
        include_dirs.insert(0, _thrust_path)
        include_dirs.insert(0, _libcudacxx_path)

    return {
        'include_dirs': include_dirs,
        'library_dirs': library_dirs,
        'define_macros': define_macros,
        'language': 'c++',
        'extra_compile_args': extra_compile_args,
    }


def _match_output_lines(output_lines, regexs):
    # Matches regular expressions `regexs` against `output_lines` and finds the
    # consecutive matching lines from `output_lines`.
    # `None` is returned if no match is found.
    if len(output_lines) < len(regexs):
        return None

    matches = [None] * len(regexs)
    for i in range(len(output_lines) - len(regexs)):
        for j in range(len(regexs)):
            m = re.match(regexs[j], output_lines[i + j])
            if not m:
                break
            matches[j] = m
        else:
            # Match found
            return matches

    # No match
    return None


def get_compiler_base_options(compiler_path: List[str]) -> List[str]:
    """Returns base options for nvcc compiler.

    """
    global _compiler_base_options
    if _compiler_base_options is None:
        _compiler_base_options = _get_compiler_base_options(compiler_path)
    return _compiler_base_options


def _get_compiler_base_options(compiler_path):
    # Try compiling a dummy code.
    # If the compilation fails, try to parse the output of compilation
    # and try to compose base options according to it.
    # compiler_path is the path to nvcc (CUDA) or hipcc (ROCm/HIP)
    with _tempdir() as temp_dir:
        test_cu_path = os.path.join(temp_dir, 'test.cu')
        test_out_path = os.path.join(temp_dir, 'test.out')
        with open(test_cu_path, 'w') as f:
            f.write('int main() { return 0; }')
        proc = subprocess.Popen(
            compiler_path + ['-o', test_out_path, test_cu_path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        stdoutdata, stderrdata = proc.communicate()
        stderrlines = stderrdata.split(b'\n')
        if proc.returncode != 0:

            # No supported host compiler
            matches = _match_output_lines(
                stderrlines,
                [
                    b'^ERROR: No supported gcc/g\\+\\+ host compiler found, '
                    b'but .* is available.$',
                    b'^ *Use \'nvcc (.*)\' to use that instead.$',
                ])
            if matches is not None:
                base_opts = matches[1].group(1)
                base_opts = base_opts.decode('utf8').split(' ')
                return base_opts

            # Unknown error
            raise RuntimeError(
                'Encountered unknown error while testing nvcc:\n' +
                stderrdata.decode('utf8'))

    return []


_hip_version = None
_thrust_version = None
_cudnn_version = None
_nccl_version = None
_cutensor_version = None
_cub_path = None
_cub_version = None
_jitify_path = None
_jitify_version = None
_compute_capabilities = None
_cusparselt_version = None


def check_hip_version(compiler, settings):
    global _hip_version
    try:
        out = build_and_run(compiler, '''
        #include <hip/hip_version.h>
        #include <stdio.h>
        int main() {
          printf("%d", HIP_VERSION);
          return 0;
        }
        ''', include_dirs=settings['include_dirs'])

    except Exception as e:
        utils.print_warning('Cannot check HIP version', str(e))
        return False

    _hip_version = int(out)

    if _hip_version < minimum_hip_version:
        utils.print_warning(
            'ROCm/HIP version is too old: %d' % _hip_version,
            'ROCm 3.5.0 or newer is required')
        return False

    return True


def get_hip_version(formatted: bool = False) -> int:
    """Return ROCm version cached in check_hip_version()."""
    global _hip_version
    if _hip_version is None:
        msg = 'check_hip_version() must be called first.'
        raise RuntimeError(msg)
    if formatted:
        return str(_hip_version)
    return _hip_version


def check_compute_capabilities(compiler, settings):
    """Return compute capabilities of the installed devices."""
    global _compute_capabilities
    try:
        src = '''
        #include <cuda_runtime_api.h>
        #include <stdio.h>
        #define CHECK_CUDART(x) { if ((x) != cudaSuccess) return 1; }

        int main() {
          int device_count;
          CHECK_CUDART(cudaGetDeviceCount(&device_count));
          for (int i = 0; i < device_count; i++) {
              cudaDeviceProp prop;
              CHECK_CUDART(cudaGetDeviceProperties(&prop, i));
              printf("%d%d ", prop.major, prop.minor);
          }
          return 0;
        }
        '''
        out = build_and_run(
            compiler, src,
            include_dirs=settings['include_dirs'],
            libraries=('cudart',),
            library_dirs=settings['library_dirs'])
        _compute_capabilities = set([int(o) for o in out.split()])
    except Exception as e:
        utils.print_warning('Cannot check compute capability\n{0}'.format(e))
        return False

    return True


def get_compute_capabilities(formatted: bool = False) -> Set[int]:
    return _compute_capabilities


def check_thrust_version(compiler, settings):
    global _thrust_version

    try:
        out = build_and_run(compiler, '''
        #include <thrust/version.h>
        #include <stdio.h>

        int main() {
          printf("%d", THRUST_VERSION);
          return 0;
        }
        ''', include_dirs=settings['include_dirs'])
    except Exception as e:
        utils.print_warning('Cannot check Thrust version\n{0}'.format(e))
        return False

    _thrust_version = int(out)

    return True


def get_thrust_version(formatted=False):
    """Return Thrust version cached in check_thrust_version()."""
    global _thrust_version
    if _thrust_version is None:
        msg = 'check_thrust_version() must be called first.'
        raise RuntimeError(msg)
    if formatted:
        return str(_thrust_version)
    return _thrust_version


def check_cudnn_version(compiler, settings):
    global _cudnn_version
    try:
        out = build_and_run(compiler, '''
        #include <cudnn.h>
        #include <stdio.h>
        int main() {
          printf("%d", CUDNN_VERSION);
          return 0;
        }
        ''', include_dirs=settings['include_dirs'])

    except Exception as e:
        utils.print_warning('Cannot check cuDNN version\n{0}'.format(e))
        return False

    _cudnn_version = int(out)

    if _cudnn_version < minimum_cudnn_version:
        min_major = str(minimum_cudnn_version)
        utils.print_warning(
            'Unsupported cuDNN version: {}'.format(str(_cudnn_version)),
            'cuDNN >=v{} is required'.format(min_major))
        return False

    if _cudnn_version > maximum_cudnn_version:
        max_major = str(maximum_cudnn_version)
        utils.print_warning(
            'Unsupported cuDNN version: {}'.format(str(_cudnn_version)),
            'cuDNN <=v{} is required'.format(max_major))
        return False

    return True


def get_cudnn_version(formatted=False):
    """Return cuDNN version cached in check_cudnn_version()."""
    global _cudnn_version
    if _cudnn_version is None:
        msg = 'check_cudnn_version() must be called first.'
        raise RuntimeError(msg)
    if formatted:
        return str(_cudnn_version)
    return _cudnn_version


def check_nccl_version(compiler, settings):
    global _nccl_version

    # NCCL 1.x does not provide version information.
    try:
        out = build_and_run(compiler,
                            '''
                            #ifndef CUPY_USE_HIP
                            #include <nccl.h>
                            #else
                            #include <rccl.h>
                            #endif
                            #include <stdio.h>
                            #ifdef NCCL_MAJOR
                            #ifndef NCCL_VERSION_CODE
                            #  define NCCL_VERSION_CODE \
                            (NCCL_MAJOR * 1000 + NCCL_MINOR * 100 + NCCL_PATCH)
                            #endif
                            #else
                            #  define NCCL_VERSION_CODE 0
                            #endif
                            int main() {
                              printf("%d", NCCL_VERSION_CODE);
                              return 0;
                            }
                            ''',
                            include_dirs=settings['include_dirs'],
                            define_macros=settings['define_macros'])

    except Exception as e:
        utils.print_warning('Cannot include NCCL\n{0}'.format(e))
        return False

    _nccl_version = int(out)

    return True


def get_nccl_version(formatted=False):
    """Return NCCL version cached in check_nccl_version()."""
    global _nccl_version
    if _nccl_version is None:
        msg = 'check_nccl_version() must be called first.'
        raise RuntimeError(msg)
    if formatted:
        if _nccl_version == 0:
            return '1.x'
        return str(_nccl_version)
    return _nccl_version


def check_nvtx(compiler, settings):
    if PLATFORM_WIN32:
        if _environment.get_nvtx_path() is None:
            utils.print_warning('NVTX unavailable')
            return False
    return True


def check_cub_version(compiler, settings):
    global _cub_version
    global _cub_path

    # This is guaranteed to work for any CUB source because the search
    # precedence follows that of include paths.
    # - On CUDA, CUB < 1.9.9 does not provide version.cuh and would error out
    # - On ROCm, hipCUB has the same version as rocPRIM (as of ROCm 3.5.0)
    try:
        out = build_and_run(compiler,
                            '''
                            #ifndef CUPY_USE_HIP
                            #include <cub/version.cuh>
                            #else
                            #include <hipcub/hipcub_version.hpp>
                            #endif
                            #include <stdio.h>

                            int main() {
                              #ifndef CUPY_USE_HIP
                              printf("%d", CUB_VERSION);
                              #else
                              printf("%d", HIPCUB_VERSION);
                              #endif
                              return 0;
                            }''',
                            include_dirs=settings['include_dirs'],
                            define_macros=settings['define_macros'])
    except Exception as e:
        # could be in a git submodule?
        try:
            # CuPy's bundle
            cupy_cub_include = os.path.join(
                cupy_builder.get_context().source_root,
                "third_party/cccl")
            a = subprocess.run(' '.join(['git', 'describe', '--tags']),
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                               shell=True, cwd=cupy_cub_include)
            if a.returncode == 0:
                tag = a.stdout.decode()[:-1]

                # CUB's tag convention changed after 1.8.0: "v1.9.0" -> "1.9.0"
                # In any case, we normalize it to be in line with CUB_VERSION
                if tag.startswith('v'):
                    tag = tag[1:]
                tag = tag.split('.')
                out = int(tag[0]) * 100000 + int(tag[1]) * 100
                try:
                    out += int(tag[2])
                except ValueError:
                    # there're local commits so tag is like 1.8.0-1-gdcbb288f,
                    # we add the number of commits to the version
                    local_patch = tag[2].split('-')
                    out += int(local_patch[0]) + int(local_patch[1])
            else:
                raise RuntimeError('Cannot determine CUB version from git tag'
                                   '\n{0}'.format(e))
        except Exception as e:
            utils.print_warning('Cannot determine CUB version\n{0}'.format(e))
            # 0: CUB is not built (makes no sense), -1: built with unknown ver
            out = -1

    _cub_version = int(out)
    settings['define_macros'].append(('CUPY_CUB_VERSION_CODE', _cub_version))
    return True  # we always build CUB


def get_cub_version(formatted=False):
    """Return CUB version cached in check_cub_version()."""
    global _cub_version
    if _cub_version is None:
        msg = 'check_cub_version() must be called first.'
        raise RuntimeError(msg)
    if formatted:
        if _cub_version == -1:
            return '<unknown>'
        return str(_cub_version)
    return _cub_version


def check_jitify_version(compiler, settings):
    global _jitify_version

    try:
        cupy_jitify_include = os.path.join(
            cupy_builder.get_context().source_root,
            "third_party/jitify")
        # Unfortunately Jitify does not have any identifiable name (branch,
        # tag, etc), so we must use the commit here
        a = subprocess.run(' '.join(['git', 'rev-parse', '--short', 'HEAD']),
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                           shell=True, cwd=cupy_jitify_include)
        if a.returncode == 0:
            out = a.stdout.decode()[:-1]  # unlike elsewhere, out is a str here
        else:
            raise RuntimeError('Cannot determine Jitify version from git')
    except Exception as e:
        utils.print_warning('Cannot determine Jitify version\n{}'.format(e))
        # 0: Jitify is not built (makes no sense), -1: built with unknown ver
        out = -1

    _jitify_version = out
    settings['define_macros'].append(('CUPY_JITIFY_VERSION_CODE',
                                      _jitify_version))
    return True  # we always build Jitify


def get_jitify_version(formatted=True):
    """Return Jitify version cached in check_jitify_version()."""
    global _jitify_version
    if _jitify_version is None:
        msg = 'check_jitify_version() must be called first.'
        raise RuntimeError(msg)
    if formatted:
        if _jitify_version == -1:
            return '<unknown>'
        return _jitify_version
    raise RuntimeError('Jitify version is a commit string')


def check_cutensor_version(compiler, settings):
    global _cutensor_version
    try:
        out = build_and_run(compiler, '''
        #include <cutensor.h>
        #include <stdio.h>
        #ifdef CUTENSOR_MAJOR
        #ifndef CUTENSOR_VERSION
        #define CUTENSOR_VERSION \
                (CUTENSOR_MAJOR * 1000 + CUTENSOR_MINOR * 100 + CUTENSOR_PATCH)
        #endif
        #else
        #  define CUTENSOR_VERSION 0
        #endif
        int main(int argc, char* argv[]) {
          printf("%d", CUTENSOR_VERSION);
          return 0;
        }
        ''', include_dirs=settings['include_dirs'])

    except Exception as e:
        utils.print_warning('Cannot check cuTENSOR version\n{0}'.format(e))
        return False

    _cutensor_version = int(out)

    if _cutensor_version < 2000:
        utils.print_warning(
            'Unsupported cuTENSOR version: {}'.format(_cutensor_version)
        )
        return False

    return True


def get_cutensor_version(formatted=False):
    """Return cuTENSOR version cached in check_cutensor_version()."""
    global _cutensor_version
    if _cutensor_version is None:
        msg = 'check_cutensor_version() must be called first.'
        raise RuntimeError(msg)
    return _cutensor_version


def check_cusparselt_version(compiler, settings):
    global _cusparselt_version
    try:
        out = build_and_run(compiler, '''
        #include <cusparseLt.h>
        #include <stdio.h>
        #ifndef CUSPARSELT_VERSION
        #define CUSPARSELT_VERSION 0
        #endif
        int main(int argc, char* argv[]) {
          printf("%d", CUSPARSELT_VERSION);
          return 0;
        }
        ''', include_dirs=settings['include_dirs'])

    except Exception as e:
        utils.print_warning('Cannot check cuSPARSELt version\n{0}'.format(e))
        return False

    _cusparselt_version = int(out)
    return True


def get_cusparselt_version(formatted=False):
    """Return cuSPARSELt version cached in check_cusparselt_version()."""
    global _cusparselt_version
    if _cusparselt_version is None:
        msg = 'check_cusparselt_version() must be called first.'
        raise RuntimeError(msg)
    return _cusparselt_version


def conda_get_target_name():
    out = None
    if PLATFORM_LINUX:
        plat = platform.machine()
        if plat == "aarch64":
            out = "sbsa-linux"
        else:
            out = f"{plat}-linux"
    elif PLATFORM_WIN32:
        out = 'x64'
    else:
        assert False
    logging.debug(f"{out=}")
    return out


def conda_update_dirs(include_dirs, library_dirs):
    # Note: These hacks are needed for the dependency detection stage to
    # function, because we create a fresh compiler instance that does not
    # honor CFLAGS etc set in the conda-build environment.
    include_dirs = list(include_dirs)
    library_dirs = list(library_dirs)

    if (int(os.environ.get('CONDA_BUILD_CROSS_COMPILATION', 0)) == 1):
        # If we're cross compiling, we need to generate stub files that are
        # executable in the build environment, not the target environment.
        # This assumes, however, that the build/host environments see the same
        # CUDA Toolkit.
        if os.environ.get('CONDA_OVERRIDE_CUDA', '0').startswith('12'):
            include_dirs.insert(
                0,
                f'{os.environ["BUILD_PREFIX"]}/targets/x86_64-linux/include')
            library_dirs.insert(
                0, f'{os.environ["BUILD_PREFIX"]}/targets/x86_64-linux/lib')
            library_dirs.insert(0, f'{os.environ["BUILD_PREFIX"]}/lib/stubs')
        elif os.environ.get('CONDA_OVERRIDE_CUDA', '0').startswith('11'):
            include_dirs.append('/usr/local/cuda/include')
            library_dirs.append('/usr/local/cuda/lib64/stubs')

        # for optional dependencies
        include_dirs.append(f'{os.environ["BUILD_PREFIX"]}/include')
        library_dirs.append(f'{os.environ["BUILD_PREFIX"]}/lib')

    if os.environ.get('CONDA_OVERRIDE_CUDA', '0').startswith('12'):
        if PLATFORM_LINUX:
            include_dirs.append(
                f'{os.environ["BUILD_PREFIX"]}/targets/'
                f'{conda_get_target_name()}/include')  # for crt headers
            library_dirs.append(f'{os.environ["PREFIX"]}/lib/stubs')
            # for optional dependencies
            include_dirs.append(f'{os.environ["PREFIX"]}/include')
            library_dirs.append(f'{os.environ["PREFIX"]}/lib')
        else:
            # there seems to be no stubs for windows
            # for optional dependencies
            include_dirs.append(
                f'{os.environ["LIBRARY_INC"]}')  # $PREFIX/Library/include
            library_dirs.append(
                f'{os.environ["LIBRARY_LIB"]}')  # $PREFIX/Library/lib

    return include_dirs, library_dirs


def build_shlib(compiler, source, libraries=(),
                include_dirs=(), library_dirs=(), define_macros=None,
                extra_compile_args=()):
    include_dirs, library_dirs = conda_update_dirs(include_dirs, library_dirs)
    logging.debug(include_dirs)

    with _tempdir() as temp_dir:
        fname = os.path.join(temp_dir, 'a.cpp')
        with open(fname, 'w') as f:
            f.write(source)
        objects = compiler.compile([fname], output_dir=temp_dir,
                                   include_dirs=include_dirs,
                                   macros=define_macros,
                                   extra_postargs=list(extra_compile_args))

        try:
            postargs = ['/MANIFEST'] if PLATFORM_WIN32 else []
            compiler.link_shared_lib(objects,
                                     os.path.join(temp_dir, 'a'),
                                     libraries=libraries,
                                     library_dirs=library_dirs,
                                     extra_postargs=postargs,
                                     target_lang='c++')
        except Exception as e:
            msg = 'Cannot build a stub file.\nOriginal error: {0}'.format(e)
            raise Exception(msg)


def build_and_run(compiler, source, libraries=(),
                  include_dirs=(), library_dirs=(), define_macros=None,
                  extra_compile_args=()):
    include_dirs, library_dirs = conda_update_dirs(include_dirs, library_dirs)
    logging.debug(include_dirs)

    with _tempdir() as temp_dir:
        fname = os.path.join(temp_dir, 'a.cpp')
        with open(fname, 'w') as f:
            f.write(source)

        objects = compiler.compile([fname], output_dir=temp_dir,
                                   include_dirs=include_dirs,
                                   macros=define_macros,
                                   extra_postargs=list(extra_compile_args))

        try:
            postargs = ['/MANIFEST'] if PLATFORM_WIN32 else []
            compiler.link_executable(objects,
                                     os.path.join(temp_dir, 'a'),
                                     libraries=libraries,
                                     library_dirs=library_dirs,
                                     extra_postargs=postargs,
                                     target_lang='c++')
        except Exception as e:
            msg = 'Cannot build a stub file.\nOriginal error: {0}'.format(e)
            raise Exception(msg)

        try:
            out = subprocess.check_output(os.path.join(temp_dir, 'a'))
            return out

        except Exception as e:
            msg = 'Cannot execute a stub file.\nOriginal error: {0}'.format(e)
            raise Exception(msg)
